/**
 * Copyright (c) 2013-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.registry_overlay;

import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.logging.Handler;
import java.util.logging.Level;
import java.util.logging.LogManager;
import java.util.logging.Logger;

import javax.management.MBeanServerConnection;
import javax.management.MBeanServerInvocationHandler;
import javax.management.MalformedObjectNameException;
import javax.management.ObjectName;
import javax.management.remote.JMXConnector;
import javax.management.remote.JMXConnectorFactory;
import javax.management.remote.JMXServiceURL;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.DefaultParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.ow2.petals.log.formatter.LogDataFormatter;
import org.ow2.petals.registry_overlay.api.Constants;
import org.ow2.petals.registry_overlay.configuration.PetalsRegistryOverlayCfgBuilder;
import org.ow2.petals.registry_overlay.configuration.PetalsRegistryOverlayCfgException;
import org.ow2.petals.registry_overlay.configuration.generated.Member;
import org.ow2.petals.registry_overlay.configuration.generated.Members;
import org.ow2.petals.registry_overlay.configuration.generated.PetalsRegistryOverlay;
import org.ow2.petals.registry_overlay.core.RegistryOverlayNode;
import org.ow2.petals.registry_overlay.core.StopRegistryThread;
import org.ow2.petals.registry_overlay.core.exception.ConfigurationException;
import org.ow2.petals.registry_overlay.core.exception.MemberIdentifierMissingException;
import org.ow2.petals.registry_overlay.core.exception.MissingMemberIdException;
import org.ow2.petals.registry_overlay.core.exception.MissingMembersException;
import org.ow2.petals.registry_overlay.core.exception.NotAClusterMemberException;
import org.ow2.petals.registry_overlay.core.exception.SeveralMemberIdException;
import org.ow2.petals.registry_overlay.core.mbean.AdminMBean;
import org.ow2.petals.registry_overlay.exception.ClusterCfgReadErrorException;
import org.ow2.petals.registry_overlay.exception.ClusterMemberCfgReadErrorException;
import org.ow2.petals.registry_overlay.exception.InvalidClusterCfgURLException;
import org.ow2.petals.registry_overlay.exception.InvalidClusterMemberCfgURLException;
import org.ow2.petals.registry_overlay.exception.InvalidLoggingURLException;
import org.ow2.petals.registry_overlay.exception.LoggingCfgReadErrorException;
import org.ow2.petals.registry_overlay.exception.PetalsRegistryOverlayException;
import org.ow2.petals.registry_overlay.exception.StopErrorException;

/**
 * The main entry point of Petals Registry Overlay.
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class Main {

    /**
     * Application part of the usage message
     */
    private final static String USAGE_APP = "petals-registry-overlay";

    /**
     * Header part of the usage message
     */
    private final static String USAGE_HEADER = "Petals Registry Overlay"
            + System.getProperty("line.separator");

    /**
     * Footer part of the usage message
     */
    private final static String USAGE_FOOTER = "\nWhich evolution would you like on Petals? Share it! http://www.petalslink.com/feedback"
            + System.getProperty("line.separator");

    /**
     * The short name of the option printing help on command line
     */
    private final static String HELP_SHORT_OPTION = "H";

    /**
     * The long name of the option printing help on command line
     */
    private final static String HELP_LONG_OPTION = "help";

    /**
     * The option printing help on command line
     */
    private final static Option HELP_OPTION = Option.builder(HELP_SHORT_OPTION).numberOfArgs(0)
            .longOpt(HELP_LONG_OPTION).desc("This help.").build();

    /**
     * The short name of the option configuring the Petals Registry Overlay
     * member
     */
    private final static String CFG_SHORT_OPTION = "c";

    /**
     * The long name of the option configuring the Petals Registry Overlay
     * member
     */
    private final static String CFG_LONG_OPTION = "config";

    /**
     * The option configuring the Petals Registry Overlay member
     */
    private final static Option CFG_OPTION = Option.builder(CFG_SHORT_OPTION).numberOfArgs(1).argName("config-url")
            .longOpt(CFG_LONG_OPTION).desc("URL of configuration file of the Petals Registry Overlay member to start.")
            .build();

    /**
     * The parameter to stop the local node
     */
    protected final static String COMMAND_STOP = "stop";

    /**
     * Name of the property defining the logging configuration file
     */
    protected static final String LOGGING_CFG_FILE_PROP_NAME = "petals.registry.overlay.member.logging-cfg-file";

    /**
     * Default logging configuration file, located in the same directory than
     * the member configuration file
     */
    protected static final String DEFAULT_LOGGING_CFG_FILE = "logging.properties";

    /**
     * Name of the property defining the configuration resource of the Petals
     * Registry Overlay
     */
    protected static final String CLUSTER_CFG_FILE_PROP_NAME = "petals.registry.overlay.cluster-definition";

    /**
     * Default name of the Petals Registry Overlay configuration resource,
     * located in the same directory than the member configuration file
     */
    protected static final String DEFAULT_CLUSTER_CFG_FILE_NAME = "cluster.xml";

    /**
     * Logger
     */
    private Logger logger;

    public static void main(final String[] args) {

        final Main main = new Main();
        System.exit(main.run(args));

    }

    /**
     * Get URL of the logging configuration file.
     * 
     * @param memberProperties
     *            Overlay member configuration properties
     * @param memberCfgUrl
     *            The configuration URL of the local overlay member
     * @return The URL of the logging configuration file
     * @throws InvalidLoggingURLException
     *             The URL of the logging configuration resource is invalid.
     * @see {@link #initLogging(Properties, URL)} to get more information on the
     *      logging configuration file
     */
    private final URL getLoggingProperties(final Properties memberProperties, final URL memberCfgUrl)
            throws InvalidLoggingURLException {

        final String logCfgFileStr = memberProperties.getProperty(LOGGING_CFG_FILE_PROP_NAME);
        try {
            if (logCfgFileStr != null) {
                return new URL(logCfgFileStr);
            } else {
                return new URL(memberCfgUrl, DEFAULT_LOGGING_CFG_FILE);
            }
        } catch (final MalformedURLException e) {
            throw new InvalidLoggingURLException(logCfgFileStr);
        }
    }

    /**
     * Init the logging system for its usage before reading the logging
     * configuration.
     */
    private final void initDefaultLogging() {
        this.logger = Logger.getLogger(RegistryOverlayNode.LOGGER_NAME);
        for (final Handler logHandler : this.logger.getHandlers()) {
            // note: it's important to have one instance per handler because the formatter is not thread safe
            logHandler.setFormatter(new LogDataFormatter());
        }
    }

    /**
     * <p>
     * Initialize the logging system.
     * </p>
     * <p>
     * By default, the logging system is initialized from the configuration file
     * '<code>logger.properties</code>' located in the same directory than the
     * configuration file of the Petals RegistryOverlay member. It is possible
     * to change this configuration using the property
     * {@value #LOGGING_CFG_FILE_PROP_NAME}of the Petals RegistryOverlay member
     * configuration (ie. {@value #DEFAULT_MEMBER_CFG_FILE}).
     * </p>
     * 
     * @param memberProperties
     *            Configuration properties of the local overlay member
     * @param memberCfgUrl
     *            The overlay member configuration URL
     * @throws InvalidLoggingURLException
     *             The URL of the logging configuration resource is invalid.
     * @throws LoggingCfgReadErrorException
     *             An error occurs reading the logging configuration resources
     */
    private final void initLogging(final Properties memberProperties, final URL memberCfgUrl)
            throws InvalidLoggingURLException, LoggingCfgReadErrorException {

        final URL logCfgUrl = this.getLoggingProperties(memberProperties, memberCfgUrl);
        try {
            final InputStream isLogCfg = logCfgUrl.openStream();
            if (isLogCfg == null) {
                LogManager.getLogManager().readConfiguration();
            } else {
                LogManager.getLogManager().readConfiguration(isLogCfg);
            }
        } catch (final IOException e) {
            throw new LoggingCfgReadErrorException(logCfgUrl, e);
        }
        this.logger = Logger.getLogger(RegistryOverlayNode.LOGGER_NAME);
    }

    /**
     * Read the configuration of the Petals Registry Overlay.
     * 
     * @param memberProperties
     *            Configuration properties of the local overlay member.
     * @param memberCfgUrl
     *            The configuration URL of the local overlay member
     * @return The configuration of the Petals Registry Overlay
     * @throws InvalidClusterCfgURLException
     *             The URL of the cluster configuration resource is invalid.
     * @throws ClusterCfgReadErrorException
     *             An error occurs reading the cluster configuration resources
     */
    private final PetalsRegistryOverlay readConfiguration(final Properties memberProperties,
            final URL memberCfgUrl) throws InvalidClusterCfgURLException,
            ClusterCfgReadErrorException {
        final String clusterCfgStr = memberProperties.getProperty(CLUSTER_CFG_FILE_PROP_NAME);
        try {
            final URL clusterCfgUrl;
            if (clusterCfgStr == null) {
                clusterCfgUrl = new URL(memberCfgUrl, DEFAULT_CLUSTER_CFG_FILE_NAME);
            } else {
                clusterCfgUrl = new URL(clusterCfgStr);
            }

            try {
                final InputStream isClusterCfg = clusterCfgUrl.openStream();
                try {
                    return PetalsRegistryOverlayCfgBuilder
                            .createPetalsRegistryOverlayCfg(isClusterCfg);
                } catch (final PetalsRegistryOverlayCfgException e) {
                    throw new ClusterCfgReadErrorException(clusterCfgUrl, e);
                } finally {
                    try {
                        isClusterCfg.close();
                    } catch (final IOException e) {
                        // NOP, we ignore this exception
                    }
                }
            } catch (final IOException e) {
                throw new ClusterCfgReadErrorException(clusterCfgUrl, e);
            }
        } catch (final MalformedURLException e) {
            throw new InvalidClusterCfgURLException(clusterCfgStr, e);
        }
    }

    /**
     * @return Options to parse
     */
    private Options createOptions() {
        final Options options = new Options();
        options.addOption(HELP_OPTION);
        options.addOption(CFG_OPTION);

        return options;
    }

    private void printUsage() {
        final HelpFormatter formatter = new HelpFormatter();
        formatter.printHelp(USAGE_APP, USAGE_HEADER, this.createOptions(), USAGE_FOOTER, true);
    }

    /**
     * Parse options and command line arguments
     * 
     * @param args
     *            Arguments from the command line
     * @return system return code
     */
    public int run(String[] args) {
        int exitCode = 0;

        this.initDefaultLogging();

        final CommandLineParser parser = new DefaultParser();
        try {
            // We can't use a static attribute for options because an error will
            // occurs in unit tests: an instance of options can be parsed once.
            final CommandLine cmd = parser.parse(this.createOptions(), args);

            try {

                if (cmd.hasOption(HELP_SHORT_OPTION)) {
                    printUsage();

                } else if (cmd.hasOption(CFG_SHORT_OPTION)) {
                    
                    final String cfgUrlStr = cmd.getOptionValue(CFG_SHORT_OPTION);
                    URL cfgUrl;
                    try {
                        cfgUrl = new URL(cfgUrlStr);
                    } catch (final MalformedURLException e) {
                        throw new InvalidClusterMemberCfgURLException(cfgUrlStr, e);
                    }
                    
                    final Properties memberProperties = new Properties();
                    try {
                        final InputStream isCfgUrl = cfgUrl.openStream();
                        try {
                            memberProperties.load(isCfgUrl);
                        } finally {
                            try {
                                isCfgUrl.close();
                            } catch (final IOException e) {
                                // NOP, we ignore this exception
                            }
                        }
                    } catch (final IOException e) {
                        throw new ClusterMemberCfgReadErrorException(cfgUrl, e);
                    }
                    
                    this.initLogging(memberProperties, cfgUrl);
                    
                    final PetalsRegistryOverlay petalsRegistryOverlayCfg = this.readConfiguration(memberProperties, cfgUrl);
                    
                    final String[] remainingArgs = cmd.getArgs();
                    if (remainingArgs.length > 0) {
                        if (remainingArgs[0].equalsIgnoreCase(COMMAND_STOP)) {
                            this.stopLocalNode(petalsRegistryOverlayCfg, memberProperties);
                            this.logger.info("Stop of the Petals Registry Overlay member in progress, check its logs.");
                        } else {
                            System.err.println("ERROR: Unknown option '" + remainingArgs[0] + "'\n");
                            printUsage();
                            exitCode = 1;
                        }
                    } else {
                        this.logger.info("Starting Petals Registry Overlay local member...");
                        final RegistryOverlayNode ron = new RegistryOverlayNode(petalsRegistryOverlayCfg,
                                memberProperties);
                        ron.start();
                        this.logger.info("Petals Registry Overlay local member started");

                        Runtime.getRuntime().addShutdownHook(new StopRegistryThread(ron));

                        synchronized (ron) {
                            try {
                                while (ron.isRunning()) {
                                    // wait for shutdown to happen
                                    ron.wait();
                                }
                            } catch (final InterruptedException e) {
                                this.logger.log(Level.SEVERE, "Petals Registry Overlay member interrupted !!!", e);
                                // NOP
                            }
                        }
                        this.logger.info("Petals Registry Overlay local member stopped");
                    }
                } else {
                    throw new ParseException("No local cluster member configuration provided !!");
                }
            } catch (final StopErrorException e) {
                this.logger.log(Level.SEVERE, "An error occurs stopping the cluster member.", e);
                exitCode = 2;
            } catch (final PetalsRegistryOverlayException e) {
                this.logger.log(Level.SEVERE, "An error occurs starting or stopping the cluster member.", e);
                exitCode = 2;
            } catch (final ConfigurationException e) {
                this.logger.log(Level.SEVERE, "An error about cluster configuration occurs", e);
                exitCode = 2;
            }

        } catch (final ParseException e) {
            System.err.println("ERROR: " + e.getMessage() + "\n");
            printUsage();
            exitCode = 1;
        }

        return exitCode;
    }

    /**
     * Stop the current Petals Registry Overlay node
     * 
     * @param petalsRegistryOverlayCfg
     *            The configuration of the current node
     * @param memberProperties
     *            Configuration properties of the current member
     * @throws ConfigurationException
     *             A configuration error occurs about the configuration of the current node
     * @throws StopErrorException
     *             An error occurs stooping the current node
     */
    private void stopLocalNode(final PetalsRegistryOverlay petalsRegistryOverlayCfg, final Properties memberProperties)
            throws ConfigurationException, StopErrorException {

        final String memberIdToCreate = memberProperties.getProperty(Constants.MEMBER_PROPERTY_NAME_MEMBER_ID);
        if (memberIdToCreate == null || memberIdToCreate.trim().isEmpty()) {
            throw new MemberIdentifierMissingException();
        }

        final Members members = petalsRegistryOverlayCfg.getMembers();
        if (members == null) {
            throw new MissingMembersException();
        }
        final List<Member> memberList = members.getMember();
        if (memberList == null || memberList.size() == 0) {
            throw new MissingMembersException();
        }
        String memberHost = null;
        for (final Member member : memberList) {
            final String memberId = member.getId();
            if (memberId == null || memberId.isEmpty()) {
                throw new MissingMemberIdException();
            } else if (member.getId().equals(memberIdToCreate)) {
                if (memberHost != null) {
                    throw new SeveralMemberIdException(memberIdToCreate);
                }
                memberHost = member.getValue();
                if (memberHost == null) {
                    memberHost = RegistryOverlayNode.DEFAULT_HOST;
                }
                // TODO: Add a check about the host (should be equals to the local host)
            }
        }
        if (memberHost == null) {
            throw new NotAClusterMemberException(memberIdToCreate);
        }

        // TODO: Introduce a JMX API for Petals Registry Overlay
        try {
            final JMXServiceURL url = new JMXServiceURL(RegistryOverlayNode.getJmxServiceURL(memberHost,
                    RegistryOverlayNode.getJmxPort(memberProperties)));

            final Map<String, Object> env = new HashMap<String, Object>();
            String username = memberProperties.getProperty(Constants.MEMBER_PROPERTY_NAME_JMX_USER);
            if (username == null || username.trim().isEmpty()) {
                this.logger.warning("No JMX user set. The default one will be used.");
                username = Constants.MEMBER_PROPERTY_DEFAULT_JMX_USER;
            }
            String password = memberProperties.getProperty(Constants.MEMBER_PROPERTY_NAME_JMX_PWD);
            if (password == null || password.trim().isEmpty()) {
                this.logger.warning("No JMX password set. The default one will be used.");
                password = Constants.MEMBER_PROPERTY_DEFAULT_JMX_PWD;
            }
            env.put(JMXConnector.CREDENTIALS, new String[] { username, password });

            final JMXConnector connector = JMXConnectorFactory.connect(url, env);
            try {
                final MBeanServerConnection mbeanServer = connector.getMBeanServerConnection();
                final AdminMBean adminMbean = MBeanServerInvocationHandler.newProxyInstance(mbeanServer,
                        new ObjectName("PetalsRegistryOverlay-" + memberIdToCreate + ":name=" + AdminMBean.MBEAN_NAME
                                + ",type=service"), AdminMBean.class, false);
                adminMbean.stop();
            } finally {
                connector.close();
            }

        } catch (final MalformedObjectNameException e) {
            throw new StopErrorException(e);
        } catch (final MalformedURLException e) {
            throw new StopErrorException(e);
        } catch (final IOException e) {
            throw new StopErrorException(e);
        }

    }
}
