/**
 * Copyright (c) 2013-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */

package org.ow2.petals.registry_overlay;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;

import java.io.File;
import java.lang.reflect.InvocationTargetException;
import java.net.URL;
import java.util.Properties;

import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.TemporaryFolder;
import org.ow2.petals.registry_overlay.configuration.PetalsRegistryOverlayCfgBuilder;
import org.ow2.petals.registry_overlay.configuration.generated.Credentials;
import org.ow2.petals.registry_overlay.configuration.generated.Member;
import org.ow2.petals.registry_overlay.configuration.generated.Members;
import org.ow2.petals.registry_overlay.configuration.generated.PetalsRegistryOverlay;

import com.ebmwebsourcing.easycommons.lang.reflect.ReflectionHelper;

/**
 * Unit tests of the main class {@link Main}.
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class MainTest {

    @Rule
    final public TemporaryFolder temporaryFolder = new TemporaryFolder();

    /**
     * <p>
     * Test to get the logging configuration URL when:
     * <ul>
     * <li>the property is not defined in local member properties.</li>
     * </ul>
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no error,</li>
     * <li>the returned URL is the default logging configuration file
     * {@value Main#DEFAULT_LOGGING_CFG_FILE} located in the directory
     * containing the local member configuration file.</li>
     * </p>
     */
    @Test
    public void testGetLoggingProperties_00() throws Throwable {

        final File localMemberCfg = this.temporaryFolder.newFile("my-member.properties");
        final URL localMemberCfgUrl = localMemberCfg.toURI().toURL();
        final URL expectedLoggingCfgUrl = new URL(localMemberCfgUrl, Main.DEFAULT_LOGGING_CFG_FILE);
        final Properties memberProperties = new Properties();
        assertEquals("Unexpected logging configuration URL", expectedLoggingCfgUrl,
                this.invokeGetLoggingProperties(memberProperties, localMemberCfgUrl));
    }

    private URL invokeGetLoggingProperties(final Properties memberProperties, final URL memberCfgUrl)
            throws Throwable {
        final Main main = new Main();
        try {
            return (URL) ReflectionHelper.invokePrivateMethod(main, "getLoggingProperties",
                    new Object[] { memberProperties, memberCfgUrl }, new Class[] {
                            Properties.class, URL.class });
        } catch (final InvocationTargetException e) {
            throw e.getCause();
        }
    }

    /**
     * <p>
     * Test to get the logging configuration URL when:
     * <ul>
     * <li>the property is defined in local member properties.</li>
     * </ul>
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no error,</li>
     * <li>the returned URL is the one defined in the local member properties.</li>
     * </p>
     */
    @Test
    public void testGetLoggingProperties_01() throws Throwable {

        final File loggingCfg = this.temporaryFolder.newFile("my-logging.properties");
        final File localMemberCfg = this.temporaryFolder.newFile("my-member.properties");
        final URL expectedLoggingCfgUrl = loggingCfg.toURI().toURL();
        final Properties memberProperties = new Properties();
        memberProperties.setProperty(Main.LOGGING_CFG_FILE_PROP_NAME,
                expectedLoggingCfgUrl.toString());
        assertEquals("Unexpected logging configuration URL", expectedLoggingCfgUrl,
                this.invokeGetLoggingProperties(memberProperties, localMemberCfg.toURI().toURL()));
    }

    /**
     * <p>
     * Test to read the configuration of the Petals Registry Overlay when:
     * <ul>
     * <li>the Petals Registry Overlay configuration property is not set in
     * local member properties.</li>
     * </ul>
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no error,</li>
     * <li>the returned configuration is read from the default configuration
     * file {@value Main#DEFAULT_CLUSTER_CFG_FILE_NAME} located in the directory
     * containing the local member configuration file.</li>
     * </p>
     */
    @Test
    public void testReadConfiguration_00() throws Throwable {

        // Create a complete configuration for a local member
        final String expectedMemberId = "member-1";
        final PetalsRegistryOverlay expectedClusterCfg = new PetalsRegistryOverlay();
        final Credentials credentials = new Credentials();
        expectedClusterCfg.setCredentials(credentials);
        credentials.setGroup("my-group");
        credentials.setPassword("my-password");
        final Members members = new Members();
        expectedClusterCfg.setMembers(members);
        final Member member1 = new Member();
        member1.setId(expectedMemberId);
        members.getMember().add(member1);
        final File localMemberCfgFolder = this.temporaryFolder.newFolder("member-cfg");
        final File clusterCfgFile = new File(localMemberCfgFolder,
                Main.DEFAULT_CLUSTER_CFG_FILE_NAME);
        PetalsRegistryOverlayCfgBuilder.marshallPetalsRegistryOverlayConf(expectedClusterCfg, clusterCfgFile);

        final File localMemberCfg = new File(localMemberCfgFolder, "my-member.properties");
        final URL localMemberCfgUrl = localMemberCfg.toURI().toURL();
        final Properties memberProperties = new Properties();
        assertNotNull("Unexpected cluster configuration read",
                this.readConfiguration(memberProperties, localMemberCfgUrl));
    }

    private PetalsRegistryOverlay readConfiguration(final Properties memberProperties,
            final URL memberCfgUrl)
            throws Throwable {
        final Main main = new Main();
        try {
            return (PetalsRegistryOverlay) ReflectionHelper.invokePrivateMethod(main,
                    "readConfiguration",
                    new Object[] { memberProperties, memberCfgUrl }, new Class[] {
                            Properties.class, URL.class });
        } catch (final InvocationTargetException e) {
            throw e.getCause();
        }
    }

    /**
     * <p>
     * Test to read the configuration of the Petals Registry Overlay when:
     * <ul>
     * <li>the Petals Registry Overlay configuration property is set in local
     * member properties.</li>
     * </ul>
     * </p>
     * <p>
     * Expected results:
     * <ul>
     * <li>no error,</li>
     * <li>the returned configuration is read from the default configuration
     * file {@value Main#DEFAULT_CLUSTER_CFG_FILE_NAME} located in the directory
     * containing the local member configuration file.</li>
     * </ul>
     * </p>
     */
    @Test
    public void testReadConfiguration_01() throws Throwable {

        // Create a complete configuration for a local member
        final String expectedMemberId = "member-1";
        final PetalsRegistryOverlay expectedClusterCfg = new PetalsRegistryOverlay();
        final Credentials credentials = new Credentials();
        expectedClusterCfg.setCredentials(credentials);
        credentials.setGroup("my-group");
        credentials.setPassword("my-password");
        final Members members = new Members();
        expectedClusterCfg.setMembers(members);
        final Member member1 = new Member();
        member1.setId(expectedMemberId);
        members.getMember().add(member1);
        final File localMemberCfgFolder = this.temporaryFolder.newFolder("member-cfg");
        final File clusterCfgFile = new File(localMemberCfgFolder, "my-cluster.cfg");
        PetalsRegistryOverlayCfgBuilder.marshallPetalsRegistryOverlayConf(expectedClusterCfg, clusterCfgFile);

        final File localMemberCfg = new File(localMemberCfgFolder, "my-member.properties");
        final URL localMemberCfgUrl = localMemberCfg.toURI().toURL();
        final Properties memberProperties = new Properties();
        memberProperties.setProperty(Main.CLUSTER_CFG_FILE_PROP_NAME, clusterCfgFile.toURI()
                .toURL().toString());
        assertNotNull("Unexpected cluster configuration read",
                this.readConfiguration(memberProperties, localMemberCfgUrl));
    }
}
